#include "Grid.h"

void CMGrid::Clear(void) {
			
	ASSERT(width); 
	ASSERT(height);

	for(int y=0; y<MAX_GRID_ALLOC; y++) {
		for(int x=0; x<MAX_GRID_ALLOC; x++) {
			cell[x][y].Clear();
		}
	}

}


void CMGrid::operator = (CMGrid& grid) {
	ASSERT(grid.width); 
	ASSERT(grid.height);

	height = grid.GetHeight();
	width = grid.GetWidth();

	for(int y=0; y<MAX_GRID_ALLOC; y++) {
		for(int x=0; x<MAX_GRID_ALLOC; x++) {
			cell[x][y] = grid.cell[x][y];
		}
	}
}


void CMGrid::SetLine(CMLine& line, en_LineAttr attr) {
	
	if(line.dir==RIGHT) cell[line.x][line.y].right = attr;
	else if(line.dir==DOWN) cell[line.x][line.y].down = attr;
	else ASSERT(0);

}


BOOL CMGrid::IsLineSet(CMLine& line) {
	
	if(line.dir == RIGHT) return (cell[line.x][line.y].right != UNCHECKED);
	if(line.dir == DOWN)  return (cell[line.x][line.y].down != UNCHECKED);
		

	ASSERT(NULL); // line is nor RIGHT nor DOWN!
	return FALSE;
}


void CMGrid::GetCount(st_count* count) {
	ASSERT(width); 
	ASSERT(height); 
	
	count->zeros  = 0;
	count->ones   = 0;
	count->twos   = 0;
	count->threes = 0;
	count->fours  = 0;
	
	for(int y=0; y<height; y++) {
		for(int x=0; x<width; x++) {
		
			switch(GetCheckedSides(x,y)) {
			  case 0: count->zeros++;  break;
			  case 1: count->ones++;   break;
			  case 2: count->twos++;   break;
			  case 3: count->threes++; break;
			  case 4: count->fours++;  break;
			}

		}
	}

	ASSERT(	count->zeros + 
			count->ones  +
			count->twos  +
			count->threes+
			count->fours == width*height);

}


BOOL CMGrid::AnyMakeFour() {
	st_count count;
	GetCount(& count);

	if(count.threes) return TRUE;
	else return FALSE;

}


int CMGrid::GetCheckedSides(int x, int y) {
	ASSERT(x>=0 && x<width);
	ASSERT(y>=0 && y<height);

	return
		(cell[x][y].right   != UNCHECKED ?    1:0   )+ 
		(cell[x][y].down    != UNCHECKED ?    1:0   )+ 
		(cell[x+1][y].down  != UNCHECKED ?    1:0   )+ 
		(cell[x][y+1].right != UNCHECKED ?    1:0   );
}

BOOL CMGrid::AreAllCellsChecked(void)
{
	st_count count;
	GetCount(& count);

	if(count.fours == height*width) return TRUE;

	return FALSE;
}


int CMGrid::DoCompletedCells(int current_player) {
	int count=0;

	for(int y=0; y<height; y++) {
		for(int x=0; x<width; x++) {
			if(GetCheckedSides(x,y)==4 && 
				cell[x][y].owner == UNCHECKED) {
					cell[x][y].owner = (en_LineAttr)current_player;
					count++;
			}
		}
	}

	return count;
}

void CMGrid::GetWhosWinning(st_WinningCount * wc)
{

	wc->p[0]=0;
	wc->p[1]=0;
	wc->p[2]=0;
	for(int y=0; y<height; y++) {
		for(int x=0; x<width; x++) {
		
			if(GetCheckedSides(x,y) == 4) {
				wc->p[(int)cell[x][y].owner]++;
			}

		}
	}

}

void CMGrid::GetMakeFour(CMLine *fourth_line) {
	int x0 = random(width);
	int y0 = random(height);

	int x=x0; 
	int y=y0;

	BOOL done = FALSE;

	do {
		ASSERT(x>=0 && x<width);   // out of range
		ASSERT(y>=0 && y<height);

		if(GetCheckedSides(x,y) == 3) {

			if(cell[x][y].right == UNCHECKED) *fourth_line = CMLine(x,y,RIGHT);
			else if(cell[x][y].down == UNCHECKED) *fourth_line = CMLine(x,y,DOWN);
			else if(cell[x+1][y].down == UNCHECKED) *fourth_line = CMLine(x+1,y,DOWN);
			else if(cell[x][y+1].right == UNCHECKED) *fourth_line = CMLine(x,y+1,RIGHT);
			else ASSERT(0);

			done = TRUE;
		}

		x++; 
		if(x >= width) { 
			x=0; y++;
			if(y >= height) { 
				y=0;
			}
		}

		ASSERT(done || y!=y0 || x!=x0);  // endless loop

	} while(!done);


}






