// LinesDoc.cpp : implementation of the CLinesDoc class
//

#include "stdafx.h"
#include "Lines.h"
#include <stdio.h>

#include "TryGrid.h"

#include "LinesDoc.h"
#include "LinesVie.h"

#include "MColorDl.h"
#include "DlgOpt.h"
#include <math.h>



#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

/////////////////////////////////////////////////////////////////////////////
// CLinesDoc

IMPLEMENT_DYNCREATE(CLinesDoc, CDocument)

BEGIN_MESSAGE_MAP(CLinesDoc, CDocument)
	//{{AFX_MSG_MAP(CLinesDoc)
	ON_COMMAND(ID_OPTIONS, OnOptions)
	ON_COMMAND(ID_GAME_ABORT, OnGameAbort)
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()

/////////////////////////////////////////////////////////////////////////////
// CLinesDoc construction/destruction

CLinesDoc::CLinesDoc() {
	// TODO: add one-time construction code here

	n_players = 2;
	player_type[0] = (en_PlayerType)1234;
	player_type[1] = HUMAN;
	player_type[2] = COMPUTER;
	player_type[3] = (en_PlayerType)1234;

	// Size of grid
	grid.SetWidth(8);
	grid.SetHeight(6);

}

CLinesDoc::~CLinesDoc()
{
}


   // SDI document implementation file
CLinesDoc * CLinesDoc::GetDoc()
{
	CFrameWnd * pFrame = (CFrameWnd *)(AfxGetApp()->m_pMainWnd);
	return (CLinesDoc *) pFrame->GetActiveDocument();
}



BOOL CLinesDoc::OnNewDocument() {

	static BOOL bFirstCall = TRUE;


	if (!CDocument::OnNewDocument())
		return FALSE;


//	CMenu* pMenu = AfxGetMainWnd()->GetMenu();
//	ASSERT(pMenu);
//	if(pMenu) pMenu->EnableMenuItem(ID_OPTIONS, MF_GRAYED);
	

	// Clear grid
	grid.Clear();

	current_player  = 1;

	again = 0;
	game_just_finished = FALSE;
	game_is_on = TRUE;

	m_bDecided = FALSE;

	if(bFirstCall) {
		bFirstCall=FALSE;
		current_player  = 0;
		game_is_on = FALSE;
	}
	else {
		CLinesView* pView = CLinesView::GetView();
		ASSERT(pView);
		pView->CalcSize();
	}


	// cursor computer move
	m_bGoingToLine = TRUE;

	m_lineLastSelection = CMLine(0,0,RIGHT);

	ASSERT(n_players<MAX_PLAYERS-1);

	return TRUE;
}

/////////////////////////////////////////////////////////////////////////////
// CLinesDoc serialization

void CLinesDoc::Serialize(CArchive& ar)
{
	if (ar.IsStoring()) 
	{

		ar << (WORD)0xFFAA;
		// TODO: add storing code here
		ar.Write(&grid, sizeof(CMGrid));

		ar << (BYTE)game_is_on;
		ar << (BYTE)game_just_finished;

		ar  << (WORD)n_players;
		ar  << (WORD)current_player;   // 1,2,3,4...
		ar  << (BYTE)again;

		ar  << (BYTE)player_type[0];  // human/computer
		ar  << (BYTE)player_type[1];  // human/computer
		ar  << (BYTE)player_type[2];  // human/computer
		ar  << (BYTE)player_type[3];  // human/computer



	}
	else
	{

		BYTE byte;
		WORD word;

		ar >> word; 
		if(word!=0xFFAA) {
			AfxGetMainWnd()->MessageBox("Wrong file or file is corrupted.");
			return;
		}

		//OnNewDocument();

		// TODO: add loading code here
		ar.Read(&grid, sizeof(CMGrid));

		ar >> byte; game_is_on = byte;
		ar >> byte; game_just_finished = byte;

		ar  >> word; n_players = word;
		ar  >> word; current_player = word;
		ar  >> byte; again = byte;

		ar  >> byte; player_type[0] = (en_PlayerType)byte; 
		ar  >> byte; player_type[1] = (en_PlayerType)byte; 
		ar  >> byte; player_type[2] = (en_PlayerType)byte; 
		ar  >> byte; player_type[3] = (en_PlayerType)byte; 

		m_lineLastSelection = CMLine(0,0,RIGHT);

		UpdateAllViews(NULL);

		
	}
}

/////////////////////////////////////////////////////////////////////////////
// CLinesDoc diagnostics

#ifdef _DEBUG
void CLinesDoc::AssertValid() const
{
	CDocument::AssertValid();
}

void CLinesDoc::Dump(CDumpContext& dc) const
{
	CDocument::Dump(dc);
}
#endif //_DEBUG

/////////////////////////////////////////////////////////////////////////////
// CLinesDoc commands




/*
#include <math.h>
void CLinesDoc::MoveMouse(CPoint to_point)
{
	CPoint pt;
	GetCursorPos(&pt);

	
	CPoint pt_inc;
	pt_inc.x = to_point.x - pt.x;
	pt_inc.y = to_point.y - pt.y;

	int incx=0;
	int incy=0;
	float mod = (float)sqrt(pt_inc.x*pt_inc.x + pt_inc.y*pt_inc.y);

	if(mod>.1) {
		incx = (int)((pt_inc.x/mod)*2);
		incy = (int)((pt_inc.y/mod)*2);
	}


	SetCursorPos(pt.x+incx, pt.y+incy);


}
*/

void CLinesDoc::HeartBeat(void)
{

	if(game_is_on) {

		if(player_type[current_player] == HUMAN) {
			//HumanTurnTick();
		}
		else if(player_type[current_player] == COMPUTER) {
			ComputerTurnTick();
		}	
		else ASSERT(0);		

	}
	
	
	if(game_just_finished) {
		// say who has won
		DisplayScore();

		game_is_on = 0;
		current_player = 0;
		game_just_finished = 0;

		
//		CMenu* pMenu = AfxGetMainWnd()->GetMenu();
//		ASSERT(pMenu);
//		if(pMenu) pMenu->EnableMenuItem(ID_OPTIONS, MF_ENABLED);



	}
	
}

void CLinesDoc::NextPlayer(CMLine& lineLastSelection) {
	
	CLinesView* pView = CLinesView::GetView();
	ASSERT(pView);
	
	int comp_cells = grid.DoCompletedCells(current_player);

	if(grid.AreAllCellsChecked()) game_just_finished = TRUE;

	again=1;
	if(comp_cells==0) {
		again=0;
		current_player++;
		m_lineLastSelection = lineLastSelection;
		if(current_player > n_players) current_player = 1;


	
	}
	else {
		pView->Invalidate(0);
	}
	
	
}

void CLinesDoc::ComputerTurnTick(void) {
	CLinesView* pView = CLinesView::GetView();
	ASSERT(pView);

	static CMLine chosen_line;
	

	if(!m_bDecided) {
		ComputerDecide(&chosen_line);
		m_bDecided = TRUE;
		MoveCursorInit();
		m_bGoingToLine = TRUE;
	}

	
	if(m_bGoingToLine) {

		BOOL arrived_yet = MoveCursorTick(chosen_line);
		
		if(arrived_yet) {
			grid.SetLine(chosen_line, (en_LineAttr)current_player);
			
			pView->Invalidate(0);
	
			m_bGoingToLine = FALSE;
			
		
			m_fSpiralFactorX = m_fSpiralFactorY = 0.0F;
			m_fCursorVel = 8.F;

			if(player_type[1]==COMPUTER && player_type[2]==COMPUTER) {
				m_bGoingToLine = TRUE;
				m_bDecided = FALSE;
				//pView->Invalidate(0);
				NextPlayer(m_lineLastSelection);

				//ASSERT(0);
				//AfxGetMainWnd->MessageBox("not returning to m_lineLastSelection");
			}


		}
	}
			// Return cursor to where the human left it
	else {  //else Coming back to lineLastSelection

		if(!again) {

			BOOL arrived_yet = MoveCursorTick(m_lineLastSelection);
			
			if(arrived_yet) {
				//m_bGoingToLine = TRUE;
				m_bDecided = FALSE;
				
				NextPlayer(m_lineLastSelection);
			}
		}
		else {
			m_bGoingToLine = TRUE;
			m_bDecided = FALSE;
			pView->Invalidate(0);
			NextPlayer(m_lineLastSelection);
			
		
		}
				
	}
	

}


void CLinesDoc::ComputerDecide(CMLine *chosen_line) {
	
	
	
	// STEP 1: Complete cells with three lines on (three-cell)
	if(grid.AnyMakeFour()) {
		grid.GetMakeFour(chosen_line);
		return;
	}



	CMTryGrid trying;

	// STEP 2: Search for a line which if set, doesn't make a three-cell
	trying.TryAllLines(grid);
	if(trying.AnyDontMakeThree()) { // not all possible lines make a three-cell

		trying.ChooseOneWhichDontMakeThree(chosen_line);
		return;
	}

	// STEP 3: Calc which is the least harming move.
	trying.ChooseLessHarming(grid, chosen_line);


}

BOOL CLinesDoc::MoveCursorTick(CMLine &target_line) {
	
	CLinesView* pView = CLinesView::GetView();
	ASSERT(pView);

	CPoint ptTo;
	CPoint ptCurrent;
	CRect rect;
	static float x_diff = 0.0F;
	static float y_diff = 0.0F;

	pView->GetLineCoord(target_line, &ptTo);
	pView->GetWindowRect(&rect);

	ptTo.x = ptTo.x + rect.left;
	ptTo.y = ptTo.y + rect.top;

	GetCursorPos(&ptCurrent);
	
	float fx = (float)ptCurrent.x + x_diff;
	float fy = (float)ptCurrent.y + y_diff;
	

	int dx = ptTo.x - ptCurrent.x;
	int dy = ptTo.y - ptCurrent.y;

	float incx=.0F;
	float incy=.0F;
	float mod = (float)sqrt(dx*dx + dy*dy);
	if(mod>.1) {
		incx = (dx/mod) * m_fCursorVel;
		incy = (dy/mod) * m_fCursorVel;
	}


	fx = fx + incx;    fx += -incy * m_fSpiralFactorX;
	fy = fy + incy;    fy +=  incx * m_fSpiralFactorY;

	int ix = (int) fx;
	int iy = (int) fy;

	x_diff = fx-(float)ix;
	y_diff = fy-(float)iy;
	
	SetCursorPos(ix, iy);

//	TRACE("dxy=(%i,%i)   ", dx,dy);

	if(abs(dx)<=m_fCursorVel*1.6 && abs(dy)<=m_fCursorVel*1.6) {
//	TRACE("dx<=2 && dy<=2");
//		AfxGetMainWnd()->MessageBox("YA LLEGE!");
		return TRUE;
	}

	return FALSE;

}

void CLinesDoc::MoveCursorInit(void) {

//	if(again) {
//		m_fSpiralFactorX = m_fSpiralFactorY = 0.0F;
//		m_fCursorVel = 8.0F;
//	}
//	else {
		m_fSpiralFactorX = (float)((random(200)-100)/300.0);
		m_fSpiralFactorY = (float)((random(200)-100)/300.0);
		m_fCursorVel = 2.0F;
//	}


}

void CLinesDoc::OnOptions() {
	CLinesView* pView = CLinesView::GetView();
	ASSERT(pView);

	CMDlgOptions dlg;

	dlg.m_nGridWidth = grid.GetWidth();
	dlg.m_nGridHeight = grid.GetHeight();
	dlg.m_bP1Type = (int) player_type[1];
	dlg.m_bP2Type = (int) player_type[2];
	dlg.m_bForceSquareCells = pView->m_bSquareCellView;


	if(dlg.DoModal() == IDOK) {

		BOOL bExchange = TRUE;
		BOOL bNewDoc = TRUE;

		if(game_is_on) {
			bExchange = FALSE;
			bNewDoc = FALSE;
			if(AfxGetMainWnd()->MessageBox("This will abort the current game", "Lines", MB_ICONSTOP|MB_OKCANCEL|MB_APPLMODAL)
				== IDOK) {
					bNewDoc = TRUE;
					bExchange = TRUE;
			}
		}
		
				

		if(bExchange) {
			grid.SetWidth(dlg.m_nGridWidth);
			grid.SetHeight(dlg.m_nGridHeight);
			player_type[1] = (en_PlayerType)dlg.m_bP1Type;
			player_type[2] = (en_PlayerType)dlg.m_bP2Type;
			pView->m_bSquareCellView = dlg.m_bForceSquareCells;
			UpdateAllViews(NULL);
		}

		if(bNewDoc) {
			OnNewDocument();
		}
		

	}

}

void CLinesDoc::OnGameAbort() {

	game_is_on = 0;
	grid.Clear();

	UpdateAllViews(NULL);

//	CMenu* pMenu = AfxGetMainWnd()->GetMenu();
//	ASSERT(pMenu);
//	if(pMenu) pMenu->EnableMenuItem(ID_OPTIONS, MF_GRAYED);
	
}

void CLinesDoc::DisplayScore() {

	int winner, looser;
	char str[300];
//	char str_title[50];

	st_WinningCount wincount;
	grid.GetWhosWinning(&wincount);


	if(wincount.p[1]>wincount.p[2]) { winner=1; looser=2; }
	else if(wincount.p[1]<wincount.p[2]) { winner=2; looser=1; }
	else winner = 0;

	float fVictoryRate = (abs(wincount.p[1]-wincount.p[2])) / (float)(wincount.p[1]+wincount.p[2]);


	CString strTitle;
	CString strComment;
			

// HUMAN vs HUMAN

	if(player_type[1]==HUMAN && player_type[2]==HUMAN) {

		strTitle.Format("Player %i wins", winner);

		if(winner) strComment = GetHumanLooseComment(looser, fVictoryRate);
		else strComment = "DRAW.    Everybody wins.";

	}
	else

// COMPUTER vs COMPUTER
	if(player_type[1] == COMPUTER && player_type[2] == COMPUTER) {
		if(winner) {
			strTitle.Format("COMPUTER vs COMPUTER");
			strComment.Format("As Expected, Computer WINS!   Well, it was only me playing...");
			
		} else strComment = "DRAW    Everybody wins";

	} 
	else {
		
// HUMAN vs COMPUTER
		if(winner) {
			strTitle.Format("Player %i wins", winner);
			// HUMAN wins
			if(player_type[winner] == HUMAN) {
				strComment.Format("Congratulations Player %i. Now you can tell your friends.", winner);
			}
	
			// COMPUTER wins
			else {
				ASSERT(player_type[looser] == HUMAN);
				if(winner) strComment = GetHumanLooseComment(looser, fVictoryRate);
			}
	
		}
		else strComment = "DRAW.    That was close!";
	}
	

	sprintf(str,
"\n\
              FINAL SCORE:\n\
\n                      Player 1:   %i                        \n\
\n                      Player 2:   %i                        \n\n\
%s",
				wincount.p[1], wincount.p[2], strComment);



	AfxGetMainWnd()->MessageBox(str, strTitle);
}


CString CLinesDoc::GetHumanLooseComment(int looser, float fVictoryRate) {

	CString strComment;
	// massive difference of scores
	if(fVictoryRate >=.8) {
		strComment.Format("Do you still want to play again Player %i ?", looser);
	}
	else
	// different scores
	if(fVictoryRate >=.5) {
		strComment.Format("Don't despair Player %i, it's only a game.", looser);
	}
	else
	// not so similar scores 
	if(fVictoryRate >=.2) {
		strComment.Format("Bad luck Player %i. Maybe next time.", looser);
	}
	else
	// similar scores
	if(fVictoryRate <=.2) {
		strComment.Format("You got very close this time Player %i", looser);
	} 
	
	else ASSERT(NULL);

	return strComment;
}
